@@ -49,6 +49,8 @@ struct _Language {
  * languages, just their names and where they are placed in the menu.
  * The available languages come from the supplied locale.alias */
 static Language languages [] = {
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Afrikaans"), "af_ZA", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Albanian"), "sq_AL", "Shqip", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
@@ -58,7 +60,9 @@ static Language languages [] = {
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Arabic (Lebanon)"), "ar_LB", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Azerbaijani"), "az_AZ", NULL, 0 },
+	{ N_("A-M|Armenian"), "hy_AM", NULL, 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Azerbaijani"), "az_AZ", "Azərbaycanca", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Basque"), "eu_ES", "Euskara", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
@@ -70,7 +74,7 @@ static Language languages [] = {
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Bulgarian"), "bg_BG", "Български", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Bosnian"), "bs_BA", NULL, 0 },
+	{ N_("A-M|Bosnian"), "bs_BA", "Bosanski", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Catalan"), "ca_ES", "Català", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
@@ -88,33 +92,47 @@ static Language languages [] = {
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|English"), "en", "English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|English (American)"), "en_US", "American English", 0 },
+	{ N_("A-M|English (USA)"), "en_US", "American English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|English (Australian)"), "en_AU", "Australian English", 0 },
+	{ N_("A-M|English (Australia)"), "en_AU", "Australian English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|English (British)"), "en_GB", "British English", 0 },
+	{ N_("A-M|English (UK)"), "en_GB", "British English", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|English (Canada)"), "en_CA", "Canadian English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|English (Ireland)"), "en_IE", "Irish English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|English (Denmark)"), "en_IE", "Danish English", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|English (South Africa)"), "en_ZA", "South African English", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Estonian"), "et_EE", "Eesti", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Finnish"), "fi_FI", "Suomi", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|French"), "fr_FR", "Français", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|French (Belgium)"), "fr_BE", "Français (Belgique)", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|French (Switzerland)"), "fr_CH", "Français (Suisse)", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Galician"), "gl_ES", "Galego", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|German"), "de_DE", "Deutsch", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Greek"), "el_GR", "ελληνικά", 0 },
+	{ N_("A-M|German (Austria)"), "de_AT", "Deutsch (Österreich)", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Gujarati"), "gu_IN", NULL, 0 },
+	{ N_("A-M|German (Switzerland)"), "de_CH", "Deutsch (Schweiz)", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Greek"), "el_GR", "Ελληνικά", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Gujarati"), "gu_IN", "ગુજરાતી", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Hebrew"), "he_IL", "עברית", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Hebrew"), "iw_IL", "עברית", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Hindi"), "hi_IN", "বাংলা", 0 },
+	{ N_("A-M|Hindi"), "hi_IN", "हिंदी", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
 	{ N_("A-M|Hungarian"), "hu_HU", "Magyar", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
@@ -148,6 +166,8 @@ static Language languages [] = {
         /*Note translate the A-M to the A-M you used in the group label */
         { N_("A-M|Mongolian"), "mn_MN", "Монгол", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Northern Sotho"), "nso_ZA", "Sesotho sa Leboa", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Norwegian (bokmal)"), "no_NO", "Norsk (bokmål)", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Norwegian (nynorsk)"), "nn_NO", "Norsk (nynorsk)", 0 },
@@ -168,7 +188,13 @@ static Language languages [] = {
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Russian"), "ru_RU", "русский", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Serbian"), "sr_YU", "", 0 },
+	{ N_("N-Z|Serbian"), "sr_YU", "српски", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Serbian"), "sr_CS", "српски", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Serbian (Latin)"), "sr_CS@Latn", "srpski (latinica)", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Serbian (Jekavian)"), "sr_CS@ije", "српски (ијекавски)", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Slovak"), "sk_SK", "Slovenský", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
@@ -199,6 +225,8 @@ static Language languages [] = {
 	{ N_("N-Z|Welsh"), "cy_GB", "Cymraeg", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
 	{ N_("N-Z|Yiddish"), "yi", "ייִדיש", 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Zulu"), "zu_ZA", "IsiZulu", 0 },
 	/* This is the POSIX/C locale for english, should really be in Other */
 	{ N_("Other|POSIX/C English"), "C", "POSIX/C English", 0 },
 	{ NULL, NULL }
@@ -240,7 +268,28 @@ find_lang (const char *language, gboolean *clean)
 
 	p = strrchr (name, '@');
 	if (p != NULL) {
+		char *mod = p+1;
 		*p = '\0';
+
+		/* attempt lookup without encoding but with the
+		   modifier first */
+		p = strrchr (name, '.');
+		if (p != NULL) {
+			char *noenc;
+			*p = '\0';
+
+			noenc = g_strconcat (name, "@", mod, NULL);
+			lang = g_hash_table_lookup (lang_names, noenc);
+			if (lang != NULL) {
+				g_free (name);
+				g_free (noenc);
+				return lang;
+			}
+			g_free (noenc);
+
+			*p = '.';
+		}
+
 		lang = g_hash_table_lookup (lang_names, name);
 		if (lang != NULL) {
 			g_free (name);
@@ -431,11 +480,12 @@ gdm_lang_read_locale_file (const char *locale_file)
 	Language *language;
 	gboolean clean;
 	char *curlocale;
+	char *getsret;
 
 	if (locale_file == NULL)
 		return NULL;
 
-	langlist = fopen (locale_file, "r");
+	VE_IGNORE_EINTR (langlist = fopen (locale_file, "r"));
 
 	if (langlist == NULL)
 		return NULL;
@@ -444,12 +494,16 @@ gdm_lang_read_locale_file (const char *locale_file)
 
 	dupcheck = g_hash_table_new (g_str_hash, g_str_equal);
 
-	while (fgets (curline, sizeof (curline), langlist)) {
+	for (;;) {
 		char *name;
 		char *lang;
 		char **lang_list;
 		int i;
 
+		VE_IGNORE_EINTR (getsret = fgets (curline, sizeof (curline), langlist));
+		if (getsret == NULL)
+			break;
+
 		if (curline[0] <= ' ' ||
 		    curline[0] == '#')
 			continue;
@@ -522,7 +576,7 @@ gdm_lang_read_locale_file (const char *locale_file)
 
 	langs = g_list_sort (langs, lang_collate);
 
-	fclose (langlist);
+	VE_IGNORE_EINTR (fclose (langlist));
 
 	return langs;
 }
